/*
 * main.c - icebuteo module "WiFi"
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

/* include icebuteo module API */
#include <icebuteo-api.h>

/* debug */
#define DEBUG 2
#include "debug.h"

/*
 * Importat details:
 *
 *	- There is no main.
 *	- Everything is callback.
 *	- Each callback gets module id (mod_id), pointer to module's data (mod_data)
 *	  and one integer parameter.
 *	- All modules share one thread, so callback must exit as soon as
 *	  possible.
 *	- All callbacks have same prototypes.
 *	- stdin, stdout and stderr is unchanged.
 *	- Module can be loaded multiple times, so do NOT use global variables,
 *	  use mod_data structure.
 *
 */

/* module real name */
const char *mod_name = "WiFi";

#define FIFO_BUF_MAX	128

/* module data */
typedef struct {
	/* important configuration */
	int var_offset;			/* first variable offset */
	long long mod_timeout_value;	/* timeout settings (0 = never) */
	int mod_timeout_always;		/* timeout is required even if window is hidden */
	int *mod_filedescriptor_value;	/* array of filedescriptors (terminated by -1) */

	/* other variables put here */

	/* configuration */
	float update_seconds;
	char *interface;
	int var_iface;
	int var_essid;
	int var_bitrate;
	int var_linkquality_str;
	int var_sig_noise_str;
	int var_linkquality_bar;
	int var_signal_bar;
	int var_noise_bar;
} MOD_DATA;

/* macro for easy access to module's data */
#define MD	((MOD_DATA *) (*mod_data))


/*****************************************************************************
 * kernel-module data exchange functions
 *
 * All of this ones returns pointer in parameter. Return value is unused.
 * Just only return pointer, don't do anything else.
 */
int mod_get_timeout_ptr(int mod_id, const void **mod_data, long long **value, int **timeout_always) {
	*value = &(MD->mod_timeout_value);
	*timeout_always = &(MD->mod_timeout_always);
	return 0;
}
int mod_get_fd_list_ptr(int mod_id, const void **mod_data, int ***list_begin) {
	*list_begin = &(MD->mod_filedescriptor_value);
	return 0;
}

/*****************************************************************************
 * Initialize callback
 *
 * Called once after module is loaded.
 * Returns zero if success.
 */
int mod_initialize(int mod_id, const void **mod_data, int var_offset, const char *arg) {
	int i = 1;
	const int cfg_tr_len = 10;
	P_TRANS cfg_tr[] = {
		/* const. name,			id, var, type */
		{ "interface",			1, NULL, P_STRING },
		{ "update-seconds",		2, NULL, P_FLOAT },
		{ "var-bitrate",		3, NULL, P_INTEGER },
		{ "var-essid",			4, NULL, P_INTEGER },
		{ "var-iface",			5, NULL, P_INTEGER },
		{ "var-linkquality-bar",	6, NULL, P_INTEGER },
		{ "var-linkquality-str",	7, NULL, P_INTEGER },
		{ "var-noise-bar",		8, NULL, P_INTEGER },
		{ "var-sig-noise-str",		10, NULL, P_INTEGER },
		{ "var-signal-bar",		9, NULL, P_INTEGER },
		{ NULL, 0, NULL, 0 }
	};
	static P_FORMAT cfg_format[] = {
		P_CONSTANT, P_EQUAL, P_STRING | P_INTEGER | P_FLOAT, 0
	};

	DBG_MSG(1, "initialize...");

	/* alloc memory for "global" variables */
	*mod_data = malloc(sizeof(MOD_DATA));
	memset((void *)*mod_data, 0, sizeof(MOD_DATA));
	MD->mod_filedescriptor_value = NULL;

	/* save offset */
	MD->var_offset = var_offset;

	/* load configuration */
	i = 0;
	cfg_tr[i++].var = &(MD->interface);
	cfg_tr[i++].var = &(MD->update_seconds);
	cfg_tr[i++].var = &(MD->var_bitrate);
	cfg_tr[i++].var = &(MD->var_essid);
	cfg_tr[i++].var = &(MD->var_iface);
	cfg_tr[i++].var = &(MD->var_linkquality_bar);
	cfg_tr[i++].var = &(MD->var_linkquality_str);
	cfg_tr[i++].var = &(MD->var_noise_bar);
	cfg_tr[i++].var = &(MD->var_sig_noise_str);
	cfg_tr[i++].var = &(MD->var_signal_bar);
	if (!parse_file(arg, NULL, cfg_format, -1, 0, 2, cfg_tr, cfg_tr_len)) {
		return -1;
	}

	/* register timeout - first after 0.1 seconds */
	MD->mod_timeout_value = 100e3;
	MD->mod_timeout_always = 0;

	/* init vars */
	var_set_as_int(MD->var_offset + 0, 1);
	var_set_as_string(MD->var_offset + MD->var_iface, MD->interface);
	var_set_as_string(MD->var_offset + MD->var_essid, "(none)");
	var_set_as_string(MD->var_offset + MD->var_bitrate, "0 Mbps");
	var_set_as_float(MD->var_offset + MD->var_linkquality_bar, 1.0);
	var_set_as_string(MD->var_offset + MD->var_sig_noise_str, "0 / 0 dBm");
	var_set_as_float(MD->var_offset + MD->var_signal_bar, 1.0);
	var_set_as_float(MD->var_offset + MD->var_noise_bar, 1.0);

	return 0;
}


/*****************************************************************************
 * Cleanup callback
 *
 * Called once before module is unloaded.
 * Parameter zero is unused (always zero).
 */
int mod_cleanup(int mod_id, const void **mod_data) {
	DBG_MSG(1, "cleanup...");
	free(MD);
	if(MD->mod_filedescriptor_value)
		free(MD->mod_filedescriptor_value);
	return 0;
}


/*****************************************************************************
 * Timeout callback
 */
int mod_timeout(int mod_id, const void **mod_data) {
	FILE *wireless;
	char line[255];
	char iface[6];
	char status [3];
	float link = 0;
	float level = 0;
	float noise = 0;
	int nwid = 0;
	int crypt = 0;
	int misc = 0;
	int mode = 0;

	MD->mod_timeout_value = MD->update_seconds * 1e6;

	DBG_MSG(1, "Updating WiFi info ...");
	if ((wireless = fopen("/proc/net/wireless", "r")) != NULL) {
		fgets(line, sizeof(line), wireless);
		fgets(line, sizeof(line), wireless);
		if (fgets(line, sizeof(line), wireless) == NULL) {
			mode = 0;
		} else {
			sscanf(line, "%s %s %f %f %f %d %d %d",
			       iface, status, &link, &level, &noise, &nwid, &crypt,
			       &misc);
			iface[strlen(iface) - 1] = 0;	/* remove ':' */
			/*
			if ((!iw_getinf_range(iface, &range))
			    || (!range.max_qual.qual))
				range.max_qual.qual = 100;
			*/
			mode = 1;
		}
		fclose(wireless);

		var_set_as_string(MD->var_offset + MD->var_iface, MD->interface);
		var_set_as_string(MD->var_offset + MD->var_essid, "?");
		var_set_as_string(MD->var_offset + MD->var_bitrate, "? Mbps");
		var_set_as_float(MD->var_offset + MD->var_linkquality_bar, link/70.0);
		var_set_as_string(MD->var_offset + MD->var_sig_noise_str, "0 / 0 dBm");
		var_set_as_float(MD->var_offset + MD->var_signal_bar, 1 + level / 100.0);
		var_set_as_float(MD->var_offset + MD->var_noise_bar, 1 + noise / 100.0);
	}
	DBG_MSG(2, "Updating WiFi info ... Done.");
	return 0;
}


/*****************************************************************************
 * filedescriptor callback
 */
int mod_filedescriptor(int mod_id, const void **mod_data, int fd) {
	DBG_MSG(1, "File descriptor callback...");
	DBG_MSG(2, "File descriptor callback... Done.");
	return 0;
}

